/* db_vm.h - definitions for a simple virtual machine
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#ifndef __DB_VM_H__
#define __DB_VM_H__

#include <stdio.h>
#include <stdarg.h>
#include <setjmp.h>
#include "db_image.h"

/* forward type declarations */
typedef struct Interpreter Interpreter;

/* intrinsic function handler type */
typedef void IntrinsicFcn(Interpreter *i);

/* interpreter state structure */
struct Interpreter {
    ImageHdr *image;
    jmp_buf errorTarget;
    VMVALUE *stack;
    VMVALUE *stackTop;
    VMVALUE *fp;
    VMVALUE *sp;
    uint8_t *pc;
    int argc;
};

/* stack manipulation macros */
#define Reserve(i, n)   do {                                    \
                            if ((i)->sp + (n) > (i)->stackTop)  \
                                StackOverflow(i);               \
                            else  {                             \
                                int cnt = (n);                  \
                                while (--cnt >= 0)              \
                                    Push(i, 0);                 \
                            }                                   \
                        } while (0)
#define CPush(i, v)     do {                                    \
                            if ((i)->sp + 1 >= (i)->stackTop)   \
                                StackOverflow(i);               \
                            else                                \
                                Push(i, v);                     \
                        } while (0)
#define Push(i, v)      (*++(i)->sp = (v))
#define Pop(i)          (*(i)->sp--)
#define Drop(i, n)      ((i)->sp -= (n))

/* prototypes */
void StackOverflow(Interpreter *i);

/* prototypes and variables from db_vmfcn.c */
extern IntrinsicFcn * FLASH_SPACE Intrinsics[];
extern int IntrinsicCount;

/* prototypes from db_vmint.c */
uint8_t *InitInterpreter(Interpreter *i, size_t stackSize);
int Execute(Interpreter *i, ImageHdr *image);
void Warn(const char *fmt, ...);                    /* fmt in FLASH_SPACE */
void Abort(Interpreter *i, const char *fmt, ...);   /* fmt in FLASH_SPACE */

/* prototypes from db_vmdebug.c */
void DecodeFunction(VMUVALUE base, const uint8_t *code, int len);
int DecodeInstruction(VMUVALUE base, const uint8_t *code, const uint8_t *lc);
void ShowStack(Interpreter *i);

void VM_sysinit(int argc, char *argv[]);
void VM_printf(const char *fmt, ...);           /* fmt in FLASH_SPACE */
void VM_vprintf(const char *fmt, va_list ap);   /* fmt in FLASH_SPACE */
void VM_putchar(int ch);
void VM_getline(char *buf, int size);

#endif
